;External assembly language equivalent for the inline assembler code in dotfloatfa.cpp
;Build under Win32: nasm -f win32 dotasm.asm

                [BITS 32]

		segment	.text

		global	_DotXMM1Acc4E
		global	_DotXMM2Acc8E
		global	_DotXMM2Acc16E

;DotXMM1Acc
;Purpose:	Compute the dot product with SSE instructions.
;Input:		[esp+4]  = pointer to X
;		[esp+8]  = pointer to Y
;		[esp+12] = n (# of array elements)
;Output:	Dot product
_DotXMM1Acc4E:

		mov     ecx,[esp+12]				;ecx = n
		mov     eax,[esp+4]				;eax -> X
		mov     edx,[esp+8]				;edx -> Y
		pxor	xmm7,xmm7				;xmm7 = accumulator for the 4-component vector
		shl     ecx,2					;ecx = 4*n (float)
		sub     esp,byte 4				;stack space for function result
		sub	eax,edx					;saves 1 lea instruction inside the main loop

		ALIGN 16

.loop:

		movaps  xmm0,[eax+edx]				;xmm0 = X[i+3] X[i+2] X[i+1] X[i+0]
		mulps	xmm0,[edx]				;xmm0 = X[i+3]*Y[i+3] X[i+2]*Y[i+2] X[i+1]*Y[i+1] X[i]*Y[i]
		addps   xmm7,xmm0				;sum up
		lea     edx,[edx+16]				;update pointers
		sub     ecx,byte 16				;count down
		jnz     .loop
		movhlps xmm0,xmm7				;get bits 64 - 127 from xmm7
		addps   xmm7,xmm0				;sums in 2 dwords
		pshufd  xmm0,xmm7,1				;get bits 32 - 63 from xmm7
		addss   xmm7,xmm0				;sum in 1 dword
		movss   [esp],xmm7				;store sum
		fld     dword [esp]				;load function result
		add     esp,byte 4				;adjust stack
		ret

;DotXMM2Acc8E
;Purpose:	Compute the dot product with SSE instructions.
;Input:		[esp+4]  = pointer to X
;		[esp+8]  = pointer to Y
;		[esp+12] = n (# of array elements)
;Output:	Dot product
_DotXMM2Acc8E:

		pxor    xmm7,xmm7				;xmm7 = 1. accumulator
		mov     ecx,[esp+12]				;ecx = n
		mov     eax,[esp+4]				;eax -> X
		mov     edx,[esp+8]				;edx -> Y
		pxor    xmm6,xmm6				;xmm6 = 2. accumulator
		shl     ecx,2					;ecx = 4*n (float)
		sub     esp,4					;stack space for function result
		sub	eax,edx					;saves 1 lea instruction inside the main loop

		ALIGN 16

.loop:

		movaps  xmm0,[eax+edx]				;xmm0 = X[i+3] X[i+2] X[i+1] X[i]
		mulps   xmm0,[edx]				;xmm0 = X[i+3]*Y[i+3] X[i+2]*Y[i+2] X[i+1]*Y[i+1] X[i]*Y[i]
		addps   xmm7,xmm0				;sum up
		movaps  xmm1,[eax+edx+16]			;xmm1 = X[i+7] X[i+6] X[i+5] X[i+4]
		mulps   xmm1,[edx+16]				;xmm1 = X[i+7]*Y[i+7] X[i+6]*Y[i+6] X[i+5]*Y[i+5] X[i+4]*Y[i+4]
		addps   xmm6,xmm1				;sum up
		lea     edx,[edx+32]				;update pointers
		sub     ecx,byte 32				;count down
		jnz     .loop
		addps   xmm7,xmm6				;add both partial vectors
		movhlps xmm0,xmm7				;get bits 64 - 127 from xmm7
		addps   xmm7,xmm0				;sums in 2 dwords
		pshufd  xmm0,xmm7,1				;get bits 32 - 63 from xmm7
		addss   xmm7,xmm0				;sum in 1 dword
		movss   [esp],xmm7				;store sum
		fld     dword [esp]				;load function result
		add     esp,byte 4				;adjust stack
		ret


;DotXMM2Acc16E
;Purpose:	Compute the dot product with SSE instructions.
;Input:		[esp+4]  = pointer to X
;		[esp+8]  = pointer to Y
;		[esp+12] = n (# of array elements)
;Output:	Dot product
_DotXMM2Acc16E:

		pxor    xmm4, xmm4				;sums are in xmm4 and xmm5
		mov     ecx,[esp+12]				;ecx = n
		mov     eax,[esp+4]				;eax -> X
		mov     edx,[esp+8]				;edx -> Y
		pxor    xmm5,xmm5
		shl     ecx,2					;ecx = 4*n (float)
		sub     esp,4					;stack space for function result
		sub	eax,edx					;saves 1 lea instruction inside the main loop

		;init pattern 1. iteration

		movaps  xmm0,[eax+edx]				;xmm0 = X[i+3] X[i+2] X[i+1] X[i+0]
		movaps  xmm1,[eax+edx+16]			;xmm1 = X[i+7] X[i+6] X[i+5] X[i+4]
		mulps   xmm0,[edx]				;xmm0 = X[i+3]*Y[i+3] X[i+2]*Y[i+2] X[i+1]*Y[i+1] X[i+0]*Y[i+0]
		sub     ecx,byte 64				;we need n-1 iterations
		jz	.epilogue				;jump: only 16 elements

		;main loop pattern

		ALIGN 16

.loop:

		movaps  xmm2,[eax+edx+32]			;xmm2 = X[i+11] X[i+10] X[i+9] X[i+8]
		mulps   xmm1,[edx+16]				;xmm1 = X[i+7]*Y[i+7] X[i+6]*Y[i+6] X[i+5]*Y[i+5] X[i+4]*Y[i+4]
		addps   xmm4,xmm0				;sum up
		movaps  xmm3,[eax+edx+48]			;xmm3 = X[i+15] X[i+14] X[i+13] X[i+12]
		mulps   xmm2,[edx+32]				;xmm2 = X[i+11]*Y[i+11] X[i+10]*Y[i+10] X[i+9]*Y[i+9] X[i+8]*Y[i+8]
		addps   xmm5,xmm1				;sum up
		movaps  xmm0,[eax+edx+64]			;xmm0 = X[i+3] X[i+2] X[i+1] X[i+0]
		mulps   xmm3,[edx+48]				;xmm3 = X[i+15]*Y[i+15] X[i+14]*Y[i+14] X[i+13]*Y[i+13] X[i+12]*Y[i+12]
		lea     edx,[edx+64]				;update pointers
		addps   xmm4,xmm2				;sum up
		movaps  xmm1,[eax+edx+16]			;mm1 = X[i+7] X[i+6] X[i+5] X[i+4]
		mulps   xmm0,[edx]				;xmm0 = X[i+3]*Y[i+3] X[i+2]*Y[i+2] X[i+1]*Y[i+1] X[i+0]*Y[i+0]
		addps   xmm5,xmm3				;sum up
		sub     ecx,byte 64				;count down
		jnz	.loop

		;exit pattern last iteration

.epilogue:

		movaps  xmm2,[eax+edx+32]			;xmm2 = X[i+11] X[i+10] X[i+9] X[i+8]
		mulps   xmm1,[edx+16]				;xmm1 = X[i+7]*Y[i+7] X[i+6]*Y[i+6] X[i+5]*Y[i+5] X[i+4]*Y[i+4]
		addps   xmm4,xmm0				;sum up
		movaps  xmm3,[eax+edx+48]			;xmm3 = X[i+15] X[i+14] X[i+13] X[i+12]
		mulps   xmm2,[edx+32]				;xmm2 = X[i+11]*Y[i+11] X[i+10]*Y[i+10] X[i+9]*Y[i+9] X[i+8]*Y[i+8]
		addps   xmm5,xmm1				;sum up
		mulps   xmm3,[edx+48]				;xmm3 = X[i+15]*Y[i+15] X[i+14]*Y[i+14] X[i+13]*Y[i+13] X[i+12]*Y[i+12]
		addps   xmm4,xmm2				;sum up
		addps   xmm5,xmm3				;sum up

		;horizontal addition

		addps   xmm4,xmm5				;sum in xmm4
		movhlps xmm0,xmm4				;get bits 64 - 127 from xmm4
		addps   xmm4,xmm0				;sum in 2 dwords
		pshufd  xmm0,xmm4,1				;get bits 32 - 63 from xmm4
		addss   xmm4,xmm0				;sum in 1 dword
		movss   [esp],xmm4				;store sum
		fld     dword [esp]				;load function result
		add     esp,byte 4				;adjust stack
		ret
